import { AtRule, Declaration } from 'postcss';
import fontpieCalc from 'fontpie-calc';
import { addSlashes, getEscapedJsonUnsafe } from 'slashes';
import { tokenize, TokenType } from '@csstools/css-tokenizer';
import { parseListOfComponentValues, isTokenNode, isFunctionNode } from '@csstools/css-parser-algorithms';

/**
 * Copyright 2023 Google LLC
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

const toCssString = string => `'${addSlashes(string, {
  getEscaped: c => c === `'` ? `\\'` : getEscapedJsonUnsafe(c)
})}'`;

/**
 * Copyright 2023 Google LLC
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

const parseUrlValue = string => parseListOfComponentValues(tokenize({
  css: string
})).map(componentValue => parseUrlToken(componentValue) ?? parseUrlFunction(componentValue)).find(Boolean) ?? null;
const parseUrlToken = componentValue => {
  if (!isTokenNode(componentValue)) {
    return null;
  }
  const token = componentValue.value;
  return token[0] === TokenType.URL ? token[4].value : null;
};
const parseUrlFunction = componentValue => {
  if (!isFunctionNode(componentValue) || componentValue.getName() !== `url`) {
    return null;
  }
  const token = componentValue.tokens().find(token => token[0] === TokenType.String);
  return token[4].value;
};

/**
 * Copyright 2023 Google LLC
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

const parseFontFamilyNameValue = string => {
  const tokens = tokenize({
    css: string
  }).filter(([tokenType]) => !IGNORED_TOKEN_TYPES.has(tokenType));
  return parseFontFamilyNameString(tokens) ?? parseFontFamilyNameIdentifiers(tokens);
};
const IGNORED_TOKEN_TYPES = new Set([TokenType.Comment, TokenType.Whitespace, TokenType.EOF]);
const parseFontFamilyNameString = tokens => {
  if (tokens.length !== 1) {
    return null;
  }
  const token = tokens[0];
  return token[0] === TokenType.String ? token[4].value : null;
};
const parseFontFamilyNameIdentifiers = tokens => {
  if (tokens.length === 0) {
    return null;
  }
  const identifiers = [];
  for (const token of tokens) {
    if (token[0] !== TokenType.Ident) {
      return null;
    }
    identifiers.push(token[4].value);
  }
  return identifiers.join(` `);
};

/**
 * Copyright 2023 Google LLC
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

const plugin = options => {
  if (!options) {
    throw new Error(`Options are required`);
  }
  return {
    postcssPlugin: `postcss-fontpie`,
    AtRule: {
      'font-face': (fontFaceRule, {
        result
      }) => {
        const fontFaceDecls = getFontFaceDecls(fontFaceRule, result);
        if (!fontFaceDecls) {
          return;
        }
        const fontFaceValues = parseFontFaceValues(fontFaceDecls, options, result);
        if (!fontFaceValues) {
          return;
        }
        fontFaceRule.before(generateFallbackFontFaceRule(fontFaceRule, fontFaceDecls, fontFaceValues));
      }
    }
  };
};
plugin.postcss = true;
const getFontFaceDecls = (fontFaceRule, result) => {
  const declsByProp = new Map();
  for (const node of fontFaceRule.nodes) {
    const decl = node;
    const {
      prop
    } = decl;
    if (!FONT_FACE_DECL_PROPS.has(prop)) {
      continue;
    }
    if (declsByProp.has(prop)) {
      decl.warn(result, `Duplicate declaration`);
      return null;
    }
    declsByProp.set(prop, decl);
  }
  const src = declsByProp.get(`src`);
  if (src === undefined) {
    fontFaceRule.warn(result, `Missing src`);
    return null;
  }
  const family = declsByProp.get(`font-family`);
  if (family === undefined) {
    fontFaceRule.warn(result, `Missing font-family`);
    return null;
  }
  return {
    src,
    family,
    style: declsByProp.get(`font-style`),
    weight: declsByProp.get(`font-weight`)
  };
};
const REQUIRED_FONT_FACE_DECL_PROPS = [`src`, `font-family`];
const FONT_FACE_DECL_PROPS = new Set([...REQUIRED_FONT_FACE_DECL_PROPS, `font-style`, `font-weight`]);
const parseFontFaceValues = (fontFaceDecls, options, result) => {
  var _fontFaceDecls$style, _fontFaceDecls$weight;
  const familyAndType = parseFontFamily(fontFaceDecls.family, options, result);
  if (familyAndType === null) {
    return null;
  }
  const filename = parseFontFilename(fontFaceDecls.src, options, result);
  if (filename === null) {
    return null;
  }
  return {
    filename,
    ...familyAndType,
    style: (_fontFaceDecls$style = fontFaceDecls.style) == null ? void 0 : _fontFaceDecls$style.value,
    weight: (_fontFaceDecls$weight = fontFaceDecls.weight) == null ? void 0 : _fontFaceDecls$weight.value
  };
};
const parseFontFilename = (srcDecl, {
  srcUrlToFilename = filename => filename
}, result) => {
  const url = parseUrlValue(srcDecl.value);
  if (!url) {
    srcDecl.warn(result, `No url`);
    return null;
  }
  return srcUrlToFilename(url);
};
const parseFontFamily = (familyDecl, {
  fontTypes
}, result) => {
  const family = parseFontFamilyNameValue(familyDecl.value);
  if (!family) {
    familyDecl.warn(result, `Bad font-family`);
    return null;
  }
  if (isFallbackFontFamily(family)) {
    return null;
  }
  if (!Object.prototype.hasOwnProperty.call(fontTypes, family)) {
    familyDecl.warn(result, `Missing font type mapping: ${family}`);
    return null;
  }
  return {
    family,
    type: fontTypes[family]
  };
};
const generateFallbackFontFaceRule = (fontFaceRule, fontFaceDecls, {
  filename,
  family,
  type,
  style,
  weight
}) => {
  const calcResult = fontpieCalc(filename, {
    name: family,
    fallback: type,
    style,
    weight
  });
  if (!calcResult) {
    // eslint-disable-next-line typescript/no-throw-literal
    throw fontFaceRule.error(`fontpie error`);
  }
  const {
    fallbackFont,
    ascentOverride,
    descentOverride,
    lineGapOverride,
    sizeAdjust
  } = calcResult;
  return new AtRule({
    name: `font-face`,
    nodes: [fontFaceDecls.family.clone({
      value: toCssString(family + FALLBACK_SUFFIX)
    }), fontFaceDecls.style, fontFaceDecls.weight, fontFaceDecls.src.clone({
      value: `local(${fallbackFont})`
    }), new Declaration({
      prop: `ascent-override`,
      value: ascentOverride
    }), new Declaration({
      prop: `descent-override`,
      value: descentOverride
    }), new Declaration({
      prop: `line-gap-override`,
      value: lineGapOverride
    }), new Declaration({
      prop: `size-adjust`,
      value: sizeAdjust
    })].filter(decl => Boolean(decl)),
    source: fontFaceRule.source
  });
};
const isFallbackFontFamily = family => family.endsWith(FALLBACK_SUFFIX);
const FALLBACK_SUFFIX = ` Fallback`;

export { plugin as default };
